
"""Globally available pluggable optimizer framework.

Copyright 2008-2012 Research Foundation State University of New York 
This file is part of QUB Express.                                          

QUB Express is free software; you can redistribute it and/or modify          
it under the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, or    
(at your option) any later version.                                  

QUB Express is distributed in the hope that it will be useful,               
but WITHOUT ANY WARRANTY; without even the implied warranty of       
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
GNU General Public License for more details.                         

You should have received a copy of the GNU General Public License,   
named LICENSE.txt, in the QUB Express program directory.  If not, see        
<http://www.gnu.org/licenses/>.                                      

"""

import numpy
import random
import scipy
import scipy.optimize
import traceback
import qubx.fast.fit
import qubx.settings
import qubx.tree
import qubx.util_types

# mil_optimize
#   Method = Simplex
#   Simplex
#      max iterations = 100
#   Methods
#      Simplex = 2
#      DFP = 8

# func = lambda pars: score
# on_iter = lambda pars, grads, iterations, score: pass
# grad_func = (lambda pars, grads, dx: score) or None

g_optimizers = [] # [ {'name' : "",
                  #    'optimize' : lambda,
                  #    'defaults' : {'opt name' : val}},
                  #    'priority' : int in range(0,11) -- 0 for disabled, 1 earlier, 10 later ]
g_opt_ix = {}
OnOptimizerAdded = qubx.util_types.WeakEvent() # (name, rec)
OnOptOptions = qubx.util_types.WeakEvent() # (name) when options changed

def add_opt(name, optimize, defaults, priority=5):
    if name in g_opt_ix:
        rec = g_optimizers[g_opt_ix[name]]
    else:
        g_opt_ix[name] = len(g_optimizers)
        rec = {}
        g_optimizers.append(rec)
    rec['name'] = name
    rec['optimize'] = optimize
    rec['defaults'] = defaults
    rec['priority'] = priority
    OnOptimizerAdded(name, rec)

def del_opt(name):
    if name in g_opt_ix:
        ix = g_opt_ix[name]
        del g_optimizers[ix]
        del g_opt_ix[name]
        for i in xrange(ix, len(g_optimizers)):
            g_opt_ix[g_optimizers[i]['name']] = i

def setup_opt(profile, defaults={}, method=None):
    qubx.settings.InitSettings() # just in case
    pro = qubx.settings.SettingsMgr[profile].active
    meth = str(pro.find('Method').data) if (method is None) else method
    if not (meth in g_opt_ix):
        if ('Method' in defaults) and (defaults['Method'] in g_opt_ix):
            meth = defaults['Method']
        else:
            meth = g_optimizers[0]['name']
        pro['Method'].data = meth
    optrec = g_optimizers[g_opt_ix[meth]]
    for opt in g_optimizers:
        mpro = pro[opt['name']]
        sysdef = opt['defaults']
        for k in sysdef:
            if not mpro[k].data:
                if k in defaults:
                    mpro[k].data = defaults[k]
                else:
                    mpro[k].data = sysdef[k]
    return optrec

def setup_opts(profile, defaults={}):
    qubx.settings.InitSettings()
    pro = qubx.settings.SettingsMgr[profile].active
    methods = pro['Methods']
    opts = []
    for opt in g_optimizers:
        optname = opt['name']
        priority = qubx.tree.node_data_or_set_def(methods, optname, opt['priority'])
        if priority:
            mpro = pro[optname]
            sysdef = opt['defaults']
            for k in sysdef:
                if not mpro[k].data:
                    if k in defaults:
                        mpro[k].data = defaults[k]
                    else:
                        mpro[k].data = sysdef[k]
            opts.append( (priority, optname, opt, mpro) )
    opts.sort()
    return opts

def set_opt_priority(profile, method, priority):
    qubx.settings.InitSettings()
    pro = qubx.settings.SettingsMgr[profile].active
    pro['Methods'][method].data = priority
    OnOptOptions(method)

def node_pydata(node):
    if node.data.type == qubx.tree.QTR_TYPE_STRING:
        return str(node.data)
    elif node.data.count > 1:
        return node.data[:]
    else:
        return node.data[0]

def get_options(profile="default_optimize", options={}, method=None):
    rec = setup_opt(profile, method=method)
    pro = qubx.settings.SettingsMgr[profile].active
    meth = str(pro['Method'].data) if (method is None) else method
    pro = pro[meth]
    opts = {}
    for k in rec['defaults']:
        if k in options:
            opts[k] = options[k]
        else:
            opts[k] = node_pydata(pro[k])
    return opts

def set_options(profile="default_optimize", options={}, method=None):
    rec = setup_opt(profile)
    pro = qubx.settings.SettingsMgr[profile].active
    if not (method is None):
        meth = method
    elif 'Method' in options:
        meth = options['Method']
    else:
        meth = str(pro['Method'].data)
    pro = pro[meth]
    for k in rec['defaults']:
        if k in options:
            pro[k].data = options[k]
    OnOptOptions(meth)

def get_opt(profile="default_optimize", options={}):
    rec = setup_opt(profile)
    optf = rec['optimize']
    opts = get_options(profile, options)
    def optimize(pars, func, on_iter, grad_func=None):
        restarts = 1
        if ('max restarts' in opts) and (restarts < opts['max restarts']):
            restarts += opts['max restarts']
        while restarts:
            restarts -= 1
            score, pars, iterations, grads, Hessian = optf(opts, pars, func, on_iter, grad_func)
            if ('max iterations' in opts) and (iterations < opts['max iterations']):
                break
        return score, pars, iterations, grads, Hessian
    return optimize

def get_opts(profile="default_optimize", options={}, on_start=lambda rec, opts: None, on_end=lambda rec, opts: None):
    def optimize(pars, func, on_iter, grad_func=None):
        score, iterations, grads, Hessian = 0.0, 0, numpy.array([0.0]*len(pars)), None
        for priority, optname, rec, opts in setup_opts(profile):
            optf = rec['optimize']
            opts = get_options(profile, options, method=optname)
            restarts = 1
            if ('max restarts' in opts) and (restarts < opts['max restarts']):
                restarts += opts['max restarts']
            on_start(rec, opts)
            while restarts:
                restarts -= 1
                score, pars, iterations, grads, Hessian = optf(opts, pars, func, on_iter, grad_func)
                if ('max iterations' in opts) and (iterations < opts['max iterations']):
                    break
            on_end(rec, opts)
        return score, pars, iterations, grads, Hessian
    return optimize

def optimize(pars, func, on_iter, grad_func=None, profile="default_optimize", options={}, multi_method=False):
    return (multi_method and get_opts or get_opt)(profile, options)(numpy.array(pars), func, on_iter, grad_func)

    
def null_optimize(options, pars, func, on_iter, grad_func=None):
    # if options["whatever"]:
    grads = numpy.zeros(shape=pars.shape, dtype=pars.dtype)
    Hessian = numpy.zeros(shape=(len(pars),len(pars)), dtype=pars.dtype)
    if grad_func:
        score = grad_func(pars, grads)
    else:
        score = func(pars)
    iterations = 1
    on_iter(pars, grads, iterations, score)
    return score, pars, iterations, grads, Hessian

# add_opt("Null", null_optimize, {'whatever': True})


def num_grad(func, nx, dx=1e-6):
    def grad_func(pars, grads, dx_specific=None):
        d = dx_specific or dx
        score = func(pars)
        xpars = numpy.array(pars[:nx], dtype='float64', copy=True)
        for ix in xrange(nx):
            xpars[:nx] = pars[:nx]
            xpars[ix] += d
            xscore = func(xpars)
            grads[ix] = (xscore - score) / d
            # print ix, d, score, xscore, grads[ix]
        return score
    return grad_func

def dfp_optimize(options, pars, func, on_iter, grad_func=None):
    nx = len(pars)
    dx = options['num grad dx']
    gradf = grad_func or num_grad(func, nx, dx)
    interrupted = [False]
    out_iterations = [0]
    out_pars = numpy.array(pars)
    perturb = options['perturb']
    if perturb:
        for i,p in enumerate(pars):
            pars[i] = p*(1.0 + perturb*(random.random() - .5))
    def dfp_func(pars, grads, do_grads):
        try:
            if do_grads:
                return gradf(pars[:nx], grads, dx)
            else:
                return func(pars[:nx])
        except:
            traceback.print_exc()
            raise
    def dfp_iter(iterations, nf, ndf, ll, pars, grads):
        try:
            out_iterations[0] = iterations
            out_pars[:] = pars[:nx]
            on_iter(pars[:nx], grads[:nx], iterations, ll)
            return 0
        except:
            traceback.print_exc()
            return -1
        
    minval, errno, grads, Hessian = qubx.fast.fit.dfpmin(pars, dfp_func, dfp_iter, options['max iterations'],
                                                     options['conv LL'], options['conv grad'], options['max step'])
    return minval, out_pars, out_iterations[0], grads, Hessian

add_opt('DFP', dfp_optimize,
        {'num grad dx' : 1e-4,
         'max iterations' : 100,
         'max restarts' : 0,
         'conv LL' : 1e-6,
         'conv grad' : 1e-6,
         'max step' : 1e-1,
         'perturb' : 1e-2},
        priority=8)


def bfgs_optimize(options, pars, func, on_iter, grad_func=None):
    nx = len(pars)
    dx = options['num grad dx']
    gradf = grad_func or num_grad(func, nx, dx)
    grads = numpy.zeros(shape=pars.shape, dtype=pars.dtype)
    Hessian = numpy.zeros(shape=(len(pars),len(pars)), dtype=pars.dtype)
    iters = [0]
    score_of_pars = {}
    
    def bfgs_func(pars):
        score = func(pars)
        score_of_pars[tuple(pars[:len(grads)])] = score
        return score
    def bfgs_grad_func(pars):
        score = gradf(pars, grads, dx)
        return grads
    def bfgs_iter(pars):
        iters[0] += 1
        on_iter(pars, grads, iters[0], score_of_pars[tuple(pars[:len(grads)])])
        
    result = scipy.optimize.fmin_bfgs(bfgs_func, pars, bfgs_grad_func, gtol=options['conv grad'],
                                      maxiter=options['max iterations'],
                                      full_output=1, disp=1, retall=0, callback=bfgs_iter)
    pars, score, grads, Hessian, nf, ng, warnflag = result
    return score, pars, iters[0], grads, Hessian


add_opt('BFGS', bfgs_optimize,
        {'num grad dx' : 1e-4,
         'max iterations' : 100,
         'conv LL' : 1e-6,
         'conv grad' : 1e-6},
        priority=0)


def simplex_optimize(options, pars, func, on_iter, grad_func=None):
    grads = numpy.zeros(shape=pars.shape, dtype=pars.dtype)
    Hessian = numpy.zeros(shape=(len(pars),len(pars)), dtype=pars.dtype)
    iters = [0]
    score_of_pars = {}
    
    def simplex_func(pars):
        score = func(pars)
        score_of_pars[tuple(pars[:len(grads)])] = score
        return score
    def simplex_iter(pars):
        iters[0] += 1
        on_iter(pars, grads, iters[0], score_of_pars[tuple(pars[:len(grads)])])
        
    pars, score, iterations, nf, warnflag = \
       scipy.optimize.fmin(simplex_func, pars, maxfun=options['max f calls'],
                           xtol=options['xtol'], ftol=options['ftol'], maxiter=options['max iterations'],
                           full_output=1, disp=1, retall=0, callback=simplex_iter)
    return score, pars, iterations, grads, Hessian

add_opt('Simplex', simplex_optimize,
        {'max f calls' : 400,
         'max iterations' : 100,
         'xtol' : 1e-6,
         'ftol' : 1e-6},
        priority=2)


#def simplex_dfp_optimize(options, pars, func, on_iter, grad_func=None):
#    options['max iterations'] = options['simplex iterations']
#    options['max f calls'] = options['simplex f calls']
#    score, pars, iterations, grads, Hessian = simplex_optimize(options, pars, func, on_iter, grad_func)
#    options['max iterations'] = options['dfp iterations']
#    options['max restarts'] = options['dfp restarts']
#    return dfp_optimize(options, pars, func, on_iter, grad_func)
#
#add_opt('Simplex-DFP', simplex_dfp_optimize,
#        {'simplex f calls' : 400,
#         'simplex iterations' : 100,
#         'xtol' : 1e-6,
#         'ftol' : 1e-6,
#         'num grad dx' : 1e-4,
#         'dfp iterations' : 100,
#         'dfp restarts' : 0,
#         'conv LL' : 1e-6,
#         'conv grad' : 1e-6,
#         'max step' : 1e-1},
#        priority=0)
